//William Barratt
//Copyright 2003 William Barratt
//stegPNG.cpp
/* First iteration of steganography program for PNG files */

#include <png.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define Trace(x) {fprintf x ; fflush(stderr); fflush(stdout);}

int main()
{
  FILE * inPNG;  //file pointer for PNG file
  inPNG = fopen("C17_clouds.png", "rb"); //set file; currently hard-coded
  png_byte header[8];                    //for use later
  png_uint_32 height, width;
  int bit_depth = 0, color_type = 0;
  bool is_png = false;                   //ditto

  if(!inPNG)                             //check validity of pointer
  {
    printf("Error reading file\n");        //error message
    return -1;                           //error code
  }

  fseek(inPNG, 0, SEEK_END);             //read to end

  if(ftell(inPNG) > 1000000)             //check file size
  {
    printf("File is too large to safely read in at once -- aborting\n");
    return -5;
  }

  rewind(inPNG);                         //return file to beginning
  
  fread(header, 1, 8, inPNG);            //read 1st 8 bytes in

  is_png = !png_sig_cmp(header, 0, 8);   //check if it's PNG
  
  if(!is_png)                            //if not PNG
  {
    printf("File does not appear to be PNG\n"); //error message
    return -6;                                //error code
  }
  else printf("File appears to be PNG\n");      //if PNG, okay


  /* The following section of code is taken mostly
   * from the libpng man pages. As of March 10, 2003,
   * I don't really know what some of it does.*/
  png_structp png_ptr = png_create_read_struct
      (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if(!png_ptr)
    return -2;

  png_infop info_ptr = png_create_info_struct(png_ptr);
  if(!info_ptr)
  {
    png_destroy_read_struct(&png_ptr, NULL, NULL);
    return -3;
  
  }
  /* Taken from O'Reilly example program 
   * setjmp must be called in every function that calls a PNG-reading
   * libpng function */
  if(setjmp(png_jmpbuf(png_ptr)))
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    return -4;
  } //end O'Reilly segment
  
 /* png_infop end_info = png_create_info_struct(png_ptr);
  if(!end_info)
  {
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    return -5;
  }*/

  png_init_io(png_ptr, inPNG);           //initialize reading

  png_set_sig_bytes(png_ptr, 8);         //inform libpng we've already read

  int png_transforms = 0;                //currently not used

  png_read_info(png_ptr, info_ptr);      //read all info up to image data

  png_get_IHDR(png_ptr, info_ptr, 
          &width, &height, &bit_depth, &color_type, NULL, NULL, NULL);

 /* display_exponent == LUT_exponent * CRT_exponent */

/*uch *readpng_get_image(double display_exponent, int *pChannels, ulg *pRowbytes)
{*/

  int display_exponent = 1;
  unsigned char * image_data = NULL;
  double  gamma;
  png_uint_32  i, rowbytes;
  png_bytepp  row_pointers = NULL;


  /* setjmp() must be called in every function that calls a PNG-reading
   * libpng function */

/*  if (setjmp(png_jmpbuf(png_ptr))) {
      png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
//      return NULL;
  }*/


  /* expand palette images to RGB, low-bit-depth grayscale images to 8 bits,
   * transparency chunks to full alpha channel; strip 16-bit-per-sample
   * images to 8 bits per sample; and convert grayscale to RGB[A] */

  if (color_type == PNG_COLOR_TYPE_PALETTE)
      png_set_expand(png_ptr);
  if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
      png_set_expand(png_ptr);
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
      png_set_expand(png_ptr);
  if (bit_depth == 16)
      png_set_strip_16(png_ptr);
  if (color_type == PNG_COLOR_TYPE_GRAY ||
      color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
      png_set_gray_to_rgb(png_ptr);

  /* unlike the example in the libpng documentation, we have *no* idea where
   * this file may have come from--so if it doesn't have a file gamma, don't
   * do any correction ("do no harm") */
  if (png_get_gAMA(png_ptr, info_ptr, &gamma))
      png_set_gamma(png_ptr, display_exponent, gamma);


  /* all transformations have been registered; now update info_ptr data,
   * get rowbytes and channels, and allocate image memory */

  png_read_update_info(png_ptr, info_ptr);

  unsigned long *pRowbytes = NULL; 
  *pRowbytes = rowbytes = png_get_rowbytes(png_ptr, info_ptr);
  
  int *pChannels = NULL;
  *pChannels = (int)png_get_channels(png_ptr, info_ptr);

  if ((image_data = (unsigned char *)malloc(rowbytes*height)) == NULL) {
       png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
   //     return NULL;
  }
  if ((row_pointers = (png_bytepp)malloc(height*sizeof(png_bytep))) == NULL) {
      png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
      free(image_data);
      image_data = NULL;
     //   return NULL;
  }

  Trace((stderr, "readpng_get_image:  channels = %d, rowbytes = %ld, height = %ld\n", *pChannels, rowbytes, height));


  /* set the individual row_pointers to point at the correct offsets */

  for (i = 0;  i < height;  ++i)
      row_pointers[i] = image_data + i*rowbytes;


  /* now we can go ahead and just read the whole image */

  png_read_image(png_ptr, row_pointers);


  /* and we're done!  (png_read_end() can be omitted if no processing of
   * post-IDAT text/time/etc. is desired) */

  free(row_pointers);
  row_pointers = NULL;

  png_read_end(png_ptr, NULL);

 /*   return image_data;
}*/


/*void readpng_cleanup(int free_image_data)
{
    if (free_image_data && image_data) {
        free(image_data);
        image_data = NULL;
    }

    if (png_ptr && info_ptr) {
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
        png_ptr = NULL;
        info_ptr = NULL;
    }
} */
  
//  png_read_png(png_ptr, info_ptr, png_transforms, NULL);

//  png_bytepp row_pointers = NULL;
  
//  row_pointers = png_get_rows(png_ptr, info_ptr);
 
  //get file attributes
 /* int height = png_get_image_height(png_ptr, info_ptr);
  int width = png_get_image_width(png_ptr, info_ptr);
  int bit_depth = png_get_bit_depth(png_ptr, info_ptr);
  int color_type = png_get_color_type(png_ptr, info_ptr);
  int filter_method = png_get_filter_type(png_ptr, info_ptr);
  int compression_type = png_get_compression_type(png_ptr, info_ptr);
  int interlace_type = png_get_interlace_type(png_ptr, info_ptr);*/

/*  printf("%i\n%i\n%i\n%i\n%i\n%i\n%i\n", height, width, bit_depth, color_type, filter_method, compression_type, interlace_type);*/

//  png_bytepp row_pointers;
//  png_read_image(png_ptr, row_pointers);
  
//  png_read_end(png_ptr, info_ptr);
//  printf("%i\n", info_ptr);
//  png_destroy_info_struct(png_ptr, info_ptr);
  /* End copied section */
  
  printf("\n");
  fclose(inPNG);                         //close file pointer

  return 0;                              //clean exit
}
